from digitalguide.whatsapp.WhatsAppUpdate import WhatsAppUpdate
from digitalguide.whatsapp.generateActions import Action

def whatsapp_loop_list(client, update: WhatsAppUpdate, context,  key, value, doppelte_antwort):
    if not key in context:
        context[key] = []

    if value in context[key]:
        Action(doppelte_antwort)(client, update, context)
        return "{}_FRAGE".format(key.upper())
    else:
        context[key].append(value)

def whatsapp_loop_list_fertig(client, update: WhatsAppUpdate, context, key, answer_id_list, fertig_antwort):
    if set(answer_id_list).issubset(set(context[key])):
        Action(fertig_antwort)(client, update, context)

def whatsapp_eval_list(client, update: WhatsAppUpdate, context, answer_id_name_list, poi, response_text):
    if not poi in context:
        context[poi] = []

    response_text += "\n"

    for id, name in answer_id_name_list:
        if id in context[poi]:
            response_text += "✅ {}\n".format(name)
        else:
            response_text += "◽ {}\n".format(name)

    client.send_message(response_text, update.get_from())

whatsapp_action_functions = {"loop_list": whatsapp_loop_list,
                             "eval_list": whatsapp_eval_list,
                             "loop_list_fertig": whatsapp_loop_list_fertig
                             }



from telegram import (Update, ReplyKeyboardRemove)
from telegram.ext import (CallbackContext)
from digitalguide.pattern import JAHRESZAHL_PATTERN, KOMMAZAHL_PATTERN

import re
import warnings

def telegram_eval_jahreszahl(update: Update, context: CallbackContext, echter_wert, richtig_text, vorher_singular_text, vorher_plural_text, spaeter_singular_text, spaeter_plural_text):
    echter_wert = int(echter_wert)
    schaetzung = int(re.findall(JAHRESZAHL_PATTERN, update.message.text)[0])
    if schaetzung == echter_wert:
        update.message.reply_text(richtig_text,
                                  reply_markup=ReplyKeyboardRemove())

    differenz = schaetzung - echter_wert
    if differenz == -1:
        update.message.reply_text(spaeter_singular_text,
                                  reply_markup=ReplyKeyboardRemove())
    elif differenz < -1:
        update.message.reply_text(spaeter_plural_text.format(abs(differenz)),
                                  reply_markup=ReplyKeyboardRemove())
    elif differenz == 1:
        update.message.reply_text(vorher_singular_text,
                                  reply_markup=ReplyKeyboardRemove())
    elif differenz > 1:
        update.message.reply_text(vorher_plural_text.format(abs(differenz)),
                                  reply_markup=ReplyKeyboardRemove())

def telegram_eval_prozentzahl(update: Update, context: CallbackContext, echter_wert, richtig_text, falsch_text):
    warnings.warn("telegram_eval_prozentzahl is deprecated", DeprecationWarning)
    return telegram_eval_kommazahl(update, context, echter_wert, richtig_text, falsch_text, falsch_text, falsch_text, falsch_text)

def telegram_eval_kommazahl(update: Update, context: CallbackContext, echter_wert, richtig_text, vorher_singular_text, vorher_plural_text, spaeter_singular_text, spaeter_plural_text):
    echter_wert = float(echter_wert)
    match = re.search(
        KOMMAZAHL_PATTERN, update.message.text)

    if match:
        if match.group('vorkomma'):
            schaetzung = int(match.group('vorkomma'))
        else:
            schaetzung = 0

        if match.group('nachkomma'):
            schaetzung += float("0."+match.group('nachkomma'))

        if schaetzung == echter_wert:
            update.message.reply_text(richtig_text,
                                    reply_markup=ReplyKeyboardRemove())
        
        differenz = schaetzung - echter_wert
        if differenz == -1:
            update.message.reply_text(spaeter_singular_text,
                                    reply_markup=ReplyKeyboardRemove())
        elif differenz < -1:
            update.message.reply_text(spaeter_plural_text.format(abs(differenz)),
                                    reply_markup=ReplyKeyboardRemove())
        elif differenz == 1:
            update.message.reply_text(vorher_singular_text,
                                    reply_markup=ReplyKeyboardRemove())
        elif differenz > 1:
            update.message.reply_text(vorher_plural_text.format(abs(differenz)),
                                    reply_markup=ReplyKeyboardRemove())


def telegram_eval_laenge(update: Update, context: CallbackContext, echter_wert, richtig_text, vorher_singular_text, vorher_plural_text, spaeter_singular_text, spaeter_plural_text):
    warnings.warn("telegram_eval_laenge is deprecated", DeprecationWarning)
    return telegram_eval_kommazahl(update, context, echter_wert, richtig_text, vorher_singular_text, vorher_plural_text, spaeter_singular_text, spaeter_plural_text)

telegram_action_functions = {"eval_jahreszahl": telegram_eval_jahreszahl,
                             "eval_prozentzahl": telegram_eval_prozentzahl,
                             "eval_kommazahl": telegram_eval_kommazahl,
                             "eval_laenge": telegram_eval_laenge,
                             }



import mongoengine
import datetime

from configparser import ConfigParser

config = ConfigParser()
config.read('config.ini')

dbname = config["bot"]["bot_name"]
class TwilioUser(mongoengine.Document):
    ProfileName = mongoengine.StringField(required=True)
    WaId = mongoengine.StringField(required=True, max_length=50)
    entry_time = mongoengine.DateTimeField(default=datetime.datetime.utcnow)
    meta = {'db_alias': dbname}

class TwilioInteraction(mongoengine.Document):
    user = mongoengine.ReferenceField(TwilioUser)
    Latitude = mongoengine.StringField()
    Longitude = mongoengine.StringField()
    SmsMessageSid = mongoengine.StringField()
    NumMedia = mongoengine.StringField()
    ProfileName = mongoengine.StringField()
    SmsSid = mongoengine.StringField()
    WaId = mongoengine.StringField()
    SmsStatus = mongoengine.StringField()
    Body = mongoengine.StringField()
    To = mongoengine.StringField()
    NumSegments = mongoengine.StringField()
    MessageSid = mongoengine.StringField()
    AccountSid = mongoengine.StringField()
    From = mongoengine.StringField()
    ApiVersion = mongoengine.StringField()
    MediaContentType0 = mongoengine.StringField()
    MediaUrl0 = mongoengine.StringField()
    MessagingServiceSid = mongoengine.StringField()
    ReferralNumMedia = mongoengine.StringField()
    date = mongoengine.DateTimeField(default=datetime.datetime.utcnow)
    meta = {'db_alias': dbname}

class TwilioUserContextState(mongoengine.Document):
    WaId = mongoengine.StringField(primary_key=True)
    context = mongoengine.DictField()
    state = mongoengine.StringField()
    meta = {'db_alias': dbname}


from digitalguide.whatsapp.WhatsAppUpdate import WhatsAppUpdate


def whatsapp_default_name(client, update: WhatsAppUpdate, context):
    context["name"] = update.entry[0].changes[0].value.contacts[0].profile_name


def whatsapp_save_text_to_context(client, update: WhatsAppUpdate, context, key):
    context[key] = update.get_message_text()


def whatsapp_save_value_to_context(client, update: WhatsAppUpdate, context, key, value):
    context[key] = value


whatsapp_action_functions = {"default_name": whatsapp_default_name,
                           "save_text_to_context": whatsapp_save_text_to_context,
                           "save_value_to_context": whatsapp_save_value_to_context
                           }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This program is dedicated to the public domain under the CC0 license.

"""
First, a few callback functions are defined. Then, those functions are passed to
the Dispatcher and registered at their respective places.
Then, the bot is started and runs until we press Ctrl-C on the command line.

Usage:
Example of a bot-user conversation using ConversationHandler.
Send /start to initiate the conversation.
Press Ctrl-C on the command line or send a signal to the process to stop the
bot.
"""

import logging
import os
from collections import defaultdict
from configparser import ConfigParser

from digitalguide import (contextActions, imageActions,
                          listenfrageActions, schaetzfragenActions,
                          writeActions)

import logging.config

from digitalguide.telegram import conversationActions
logger = logging.config.fileConfig("logging.ini")
logger = logging.getLogger(__name__)

config = ConfigParser()
config.read("config.ini")

class Bot:
    def __init__(self, custom_actions={}):
        self.LOG_USER = (os.getenv('LOG_USER', 'True') == 'True')
        self.LOG_INTERAKTION = (os.getenv('LOG_INTERAKTION', 'True') == 'True')
        self.RUN_LOCAL = (os.getenv('RUN_LOCAL', 'False') == 'True')

        self.MESSANGER = os.getenv('MESSANGER', 'Telegram')

        if self.MESSANGER == "Telegram":
            from digitalguide.telegram.errorHandler import error_handler
            from digitalguide.telegram.generateActions import (callback_query_handler,
                                                        read_action_yaml)
            from digitalguide.telegram.generateStates import read_state_yml
            from telegram.ext import (CallbackQueryHandler, CommandHandler,
                                        ConversationHandler, Filters, MessageHandler,
                                        Updater)
            self.TOKEN = os.environ.get('TELEGRAM_TOKEN')
            self.PORT = int(os.environ.get('PORT', '8080'))

            #my_persistence = DBPersistence("klimafuehrung_persistencedb")
            self.updater = Updater(self.TOKEN,
                                # persistence=my_persistence,
                                use_context=True)

            self.action_functions = {**contextActions.telegram_action_functions,
                                **listenfrageActions.telegram_action_functions,
                                **conversationActions.telegram_action_functions,
                                **schaetzfragenActions.telegram_action_functions,
                                **imageActions.telegram_action_functions,
                                **writeActions.telegram_action_functions,
                                }

            self.action_functions = {**self.action_functions, **custom_actions}

            generalActions = read_action_yaml(
                "actions/telegram_general.yml", action_functions=self.action_functions, log_user=self.LOG_USER, log_interaction=self.LOG_INTERAKTION)

            cqh = callback_query_handler({**generalActions})

            prechecks = [CommandHandler('start', generalActions[config["bot"]["start_action"]]),
                            CallbackQueryHandler(cqh)]

            conv_handler = ConversationHandler(
                allow_reentry=True,
                per_chat=False,
                conversation_timeout=6 * 60 * 60,
                entry_points=[CommandHandler(
                    'start', generalActions[config["bot"]["start_action"]])],
                # persistent=True,
                name=config["bot"]["bot_name"],

                states={
                    **read_state_yml("states/telegram_general.yml", actions={**generalActions}, prechecks=prechecks),

                    ConversationHandler.TIMEOUT: [MessageHandler(Filters.regex(r'^(.)+'), generalActions["timeout"])],
                },

                fallbacks=[]
            )

            self.updater.dispatcher.add_handler(conv_handler)
            self.updater.dispatcher.add_error_handler(error_handler)


        elif self.MESSANGER == "Twilio":
            from digitalguide.twilio.generateActions import read_action_yaml
            from digitalguide.twilio.generateStates import (CommandHandler,
                                                                read_state_yml)
            from digitalguide.twilio.TwilioUpdate import TwilioUpdate
            from quart import Quart, Response, request
            from twilio.rest import Client

            # Find your Account SID and Auth Token at twilio.com/console
            # and set the environment variables. See http://twil.io/secure
            account_sid = os.environ['TWILIO_ACCOUNT_SID']
            auth_token = os.environ['TWILIO_AUTH_TOKEN']
            client = Client(account_sid, auth_token)

            self.app = Quart(__name__)

            user_states = defaultdict(lambda: "START_START")
            user_context = defaultdict(dict)

            self.action_functions = {**contextActions.twilio_action_functions,
                                **listenfrageActions.twilio_action_functions,
                                **schaetzfragenActions.twilio_action_functions,
                                **imageActions.twilio_action_functions,
                                **writeActions.twilio_action_functions,
                                }
        
            action_functions = {**self.action_functions, **custom_actions}

            general_actions = read_action_yaml("actions/twilio_general.yml", action_functions=action_functions)

            prechecks = [CommandHandler('start', general_actions[config["bot"]["start_action"]]),
                            ]

            general_handler = read_state_yml("states/twilio_general.yml", actions={
                                                **general_actions}, prechecks=prechecks)

            @self.app.route('/bot', methods=['POST'])
            def bot():
                print(request.values)
                update = TwilioUpdate(**request.values)
                icoming_state = user_states[update.From]
                context = user_context[update.From]

                print("Current context: {}".format(context))
                print("Current state: {}".format(icoming_state))
                print("Update: {}".format(str(request.values)))

                for handler in prechecks + general_handler[icoming_state]:
                    print("Filter Eval: {}".format(handler))
                    if handler.check_update(update):
                        print("Filter True: {}".format(handler))
                        new_state = handler.callback(client, update, context)
                        if new_state:
                            user_states[update.From] = new_state
                        break
                return Response(status=200)

    def run(self):
        if self.MESSANGER == "Telegram":
            if self.RUN_LOCAL:
                self.updater.start_polling()
            else:
                self.updater.start_webhook(listen="0.0.0.0",
                                        port=self.PORT,
                                        url_path=self.TOKEN,
                                        webhook_url=os.environ.get("APP_URL") + self.TOKEN)

            self.updater.idle()

        if self.MESSANGER == "Twilio":
            logger.error("Please use the create_app method.")

    def create_app(self):
        if self.MESSANGER == "Telegram":
            logger.error("Please use the run method.")
        if self.MESSANGER == "Twilio":
            return self.app


from telegram import (Update)
from telegram.ext import (CallbackContext)

from digitalguide.twilio.TwilioUpdate import TwilioUpdate


def telegram_default_name(update: Update, context: CallbackContext):
    context.user_data["name"] = update.message.from_user.first_name

def telegram_save_text_to_context(update: Update, context: CallbackContext, key):
    context.user_data[key] = update.message.text

def telegram_save_value_to_context(update: Update, context: CallbackContext, key, value):
    context.user_data[key] = value

telegram_action_functions = {"default_name": telegram_default_name,
                             "save_text_to_context": telegram_save_text_to_context,
                             "save_value_to_context": telegram_save_value_to_context
                             }



import re

# flags (iOS)
# symbols & pictographs
# emoticons
# transport & map symbols
# alchemical symbols
# Geometric Shapes Extended
# Supplemental Arrows-C
# Supplemental Symbols and Pictographs
# Chess Symbols
# Symbols and Pictographs Extended-A
# Dingbats
EMOJI_PATTERN = r"["\
    r"\U0001F1E0-\U0001F1FF"\
    r"\U0001F300-\U0001F5FF"\
    r"\U0001F600-\U0001F64F"\
    r"\U0001F680-\U0001F6FF"\
    r"\U0001F700-\U0001F77F"\
    r"\U0001F780-\U0001F7FF"\
    r"\U0001F800-\U0001F8FF"\
    r"\U0001F900-\U0001F9FF"\
    r"\U0001FA00-\U0001FA6F"\
    r"\U0001FA70-\U0001FAFF"\
    r"\U00002702-\U000027B0"\
    r"\U000024C2-\U0001F251"\
    r"]+"

JAHRESZAHL_PATTERN = r"(?P<jahreszahl>\d{1,4})"

KOMMAZAHL_PATTERN = r"(?P<vorkomma>\d+),? ?(?P<nachkomma>\d*)"

ZURUECK_PATTERN = "^("\
    "zurueck|"\
    "zurück"\
    ")$"\

WEITER_PATTERN ="^("\
    "gefunden|"\
    "bescheid|"\
    "weiter|"\
    "next|"\
    "nächster|"\
    "weit|"\
    "witer|"\
    "weitr|"\
    "überspringen|"\
    "uberspringen|"\
    "ueberspringen|"\
    "ueber springen|"\
    "waiter|"\
    "bin soweit|"\
    ")$"

WOHIN_PATTERN = "^("\
    "wohin|"\
    "wo|"\
    "weg|"\
    "wo lang|"\
    "route|"\
    "ziel|"\
    "ort|"\
    "woin|"\
    "treffpunkt|"\
    "GPS|"\
    "way|"\
    "where|"\
    "location|"\
    "Hilfe|"\
    "hilfe|"\
    "wie|"\
    "Wie"\
    ")$"

JA_PATTERN = "^("\
    "Okay|"\
    "ok|"\
    "okay|"\
    "OK|"\
    "Ok|"\
    "ja|"\
    "Ja|"\
    "Jap|"\
    "Jo|"\
    "Joa|"\
    "Yo|"\
    "Yap|"\
    "Yes|"\
    "Yess|"\
    "Yesss|"\
    "Jawohl|"\
    "jawol|"\
    "Jawoll|"\
    "Auf jeden Fall|"\
    "Auf jeden|"\
    "Klar|"\
    "Klaro|"\
    "Ci|"\
    "Cí|"\
    "Okay|"\
    "Ok|"\
    "Oki|"\
    "kay|"\
    "jes|"\
    "jep|"\
    "yep|"\
    "yop|"\
    "yup|"\
    "yupp|"\
    "Bin dabei|"\
    "dabie|"\
    "Gefunden|"\
    "bin da|"\
    "bin hier|"\
    "da|"\
    "hier|"\
    "angekommen|"\
    "geschafft|"\
    "fertig|"\
    "done|"\
    "👍|"\
    "👌|"\
    "🤚|"\
    "💪|"\
    "of course|"\
    "made it|"\
    "here|"\
    "found it|"\
    "bereit|"\
    "breit|"\
    "bin bereit|"\
    "ready|"\
    "readi|"\
    "redy|"\
    "kann losgehen|"\
    "kan losgehen|"\
    "okey|"\
    "oke|"\
    "abgemacht|"\
    "einverstanden|"\
    "ein verstanden|"\
    "gut|"\
    "gud|"\
    "jut|"\
    "ordnungsgemäß|"\
    "gebongt|"\
    "gecheckt|"\
    "gescheckt|"\
    "gechekt|"\
    "ist geritzt|"\
    "is geritzt|"\
    "all right|"\
    "allright|"\
    "d'accord|"\
    "daccord|"\
    "find ich gut|"\
    "find ich super|"\
    "sicher|"\
    "sure|"\
    "freilich|"\
    "freili|"\
    "logo|"\
    "logen|"\
    "na logo|"\
    "natürlich|"\
    "türlich|"\
    "tuerlich|"\
    "natuerlich|"\
    "immer|"\
    "alle Mal|"\
    "allemal|"\
    "gewiss|"\
    "gewiß|"\
    "gewis|"\
    "fraglos|"\
    "wahrlich|"\
    "warlich|"\
    "ausreichend|"\
    "zweifellos|"\
    "allerdings|"\
    "mit Sicherheit|"\
    "topp|"\
    "top|"\
    "oki doki|"\
    "okidoki|"\
    "von mir aus|"\
    "meinetwegen|"\
    "meinet wegen|"\
    "wenns sein muss|"\
    "wenn’s sein muss|"\
    "wens sein muss|"\
    "wen’s sein muss|"\
    "vermutlich|"\
    "wahrscheinlich|"\
    "glaube schon|"\
    "glaub schon|"\
    "bestimmt|"\
    "in der Tat|"\
    "positiv|"\
    ")$"

NEIN_PATTERN = "^("\
    "Nein|"\
    "nein|"\
    "Nope|"\
    "nop|"\
    "Nee|"\
    "ne|"\
    "neee|"\
    "nain|"\
    "nö|"\
    "auf keinen Fall|"\
    "auf keinsten|"\
    "no|"\
    "nada|"\
    "nien|"\
    "nicht|"\
    "never|"\
    "👎|"\
    "🙅‍♀️|"\
    "🙅‍♂️|"\
    "nimmermehr|"\
    "negativ|"\
    "veto|"\
    "weto|"\
    "keinesfalls|"\
    "nie und nimmer|"\
    "nieundnimmer|"\
    "sicher nicht|"\
    "unmöglich|"\
    "i wo|"\
    "mitnichten|"\
    "keineswegs|"\
    "gar nicht|"\
    ")$"


import os
import asyncio
from collections import defaultdict
from digitalguide.whatsapp.WhatsAppUpdate import WhatsAppUpdate
import yaml

from configparser import ConfigParser
config = ConfigParser()
config.read('config.ini')


def read_action_yaml(filename, action_functions={}, log_user=True, log_interaction=True):
    with open(filename) as file:
        yaml_dict = yaml.load(file, Loader=yaml.FullLoader)

    actions_dict = {}

    for key, value in yaml_dict.items():
        actions_dict[key] = Action(
            value, action_functions=action_functions, log_user=log_user, log_interaction=log_interaction)

    return actions_dict


class Action():
    def __init__(self, actions, action_functions={}, log_user=True, log_interaction=True):
        self.actions = actions
        self.action_functions = action_functions
        self.log_user = log_user
        self.log_interaction = log_interaction
        if log_user or log_interaction:
            import mongoengine
            dbname = config["bot"]["bot_name"]
            if os.getenv('DATABASE_CERT', None):
                # This can be remove in the future
                with open("ca-certificate.crt", "w") as text_file:
                    text_file.write(os.getenv('DATABASE_CERT'))
                mongoengine.connect(alias=dbname, host="mongodb+srv://" + os.getenv("DATABASE_USERNAME")+":" + os.getenv("DATABASE_PASSWORD") +
                                    "@" + os.getenv("DATABASE_HOST") + "/"+dbname+"?authSource=admin&tls=true&tlsCAFile=ca-certificate.crt")
            else:
                mongoengine.connect(alias=dbname, host="mongodb+srv://" + os.getenv("DATABASE_USERNAME")+":" + os.getenv("DATABASE_PASSWORD") +
                                    "@" + os.getenv("DATABASE_HOST") + "/"+dbname+"?authSource=admin&tls=true")

    async def __call__(self, client, update: WhatsAppUpdate, context):
        if self.log_user and not ("user_id" in context.keys()):
            from digitalguide.whatsapp.db_objects import WhatsAppUser
            db_user = WhatsAppUser(ProfileName=update.entry[0].changes[0].value.contacts[0].profile_name,
                                   WaId=update.get_from())
            db_user.save()
            context["user_id"] = db_user

        if self.log_interaction:
            from digitalguide.whatsapp.db_objects import WhatsAppInteraction

            WhatsAppInteraction(user=context["user_id"],
                                ProfileName=update.entry[0].changes[0].value.contacts[0].profile_name,
                                WaId=update.get_from(),
                                text=update.get_message_text(),
                                ).save()

        for item in self.actions:
            if item["type"] == "return":
                return item["state"]

            elif item["type"] == "message":
                placeholder_dict = {**context}
                placeholder_dict["name"] = update.entry[0].changes[0].value.contacts[0].profile_name
                placeholder_dict["echo"] = update.get_message_text()

                if "reply_buttons" in item.keys():
                    buttons = []

                    for button in item["reply_buttons"]:
                        buttons.append({
                            "type": "reply",
                            "reply": {
                                "id": button["id"],
                                "title": button["text"]
                            }
                        })

                    button_dict = {
                        "type": "button",
                                "body": {
                                    "text": item["text"].format(**placeholder_dict)
                                },
                        "action": {
                                    "buttons": buttons
                                }
                    }

                    if "footer" in item.keys():
                        button_dict["footer"] = {"text": item["footer"]}

                    client.send_reply_button(recipient_id=update.get_from(),
                                             button=button_dict)

                elif "button" in item.keys() and "section_title" in item.keys():
                    button_dict = {"body": {
                        "text": item["text"].format(**placeholder_dict)
                    }, "action": {
                        "button": item["button"],
                        "sections": [
                            {
                                "title": item["section_title"],
                                "rows": [],
                            }
                        ],
                    }
                    }

                    for row in item["rows"]:
                        row_dict = {"id": row["id"],
                                    "title": row["title"],
                                    "description": row.get("description", "")}
                        button_dict["action"]["sections"][0]["rows"].append(row_dict)

                    if "footer" in item.keys():
                        button_dict["footer"] = {"text": item["footer"]}

                    if "header" in item.keys():
                        button_dict["header"] = {"text": item["header"]}

                    print(button_dict)

                    client.send_button(
                        recipient_id=update.get_from(),
                        button=button_dict)

                else:
                    client.send_message(item["text"].format(
                        **placeholder_dict), update.get_from())

            elif item["type"] == "venue":
                client.send_location(
                    name=item["title"],
                    lat=item["latitude"],
                    long=item["longitude"],
                    address=item["address"],
                    recipient_id=update.get_from()
                )
            elif item["type"] == "photo":
                client.send_image(
                    item["url"],
                    update.get_from()
                )
            elif item["type"] == "video":
                client.send_video(
                    item["url"],
                    update.get_from()
                )
            elif item["type"] == "media_group":
                pass
                # message = client.messages.create(
                #    media_url=item["urls"],
                #    from_=update.To,
                #    to=update.From
                # )
            elif item["type"] == "audio" or item["type"] == "voice":
                client.send_audio(
                    item["url"],
                    update.get_from()
                )
            elif item["type"] == "poll":
                message = item["question"] + "\n"
                for option in item["options"]:
                    message += option + "\n"
                client.send_message(message, update.get_from())

            elif item["type"] == "function":
                arguments = {i: item[i]
                             for i in item if i != 'type' and i != 'func'}
                self.action_functions[item["func"]](
                    client, update, context, **arguments)



import os
from configparser import ConfigParser
from datetime import datetime

import boto3
import requests

from digitalguide.twilio.TwilioUpdate import TwilioUpdate

config = ConfigParser()
config.read("config.ini")

if "space" in config.keys():
    session = boto3.session.Session()
    s3_client = session.client('s3',
                            region_name=config["space"]["region_name"],
                            endpoint_url=config["space"]["endpoint_url"],
                            aws_access_key_id=os.getenv('SPACES_KEY'),
                            aws_secret_access_key=os.getenv('SPACES_SECRET'))

def twilio_write_photo(client, update: TwilioUpdate, context, bucket, folder):
    user_id = update.WaId
    name = update.ProfileName
    update.MediaUrl0

    s3_client.put_object(Bucket=bucket,
                      Key=folder + "/" +
                      str(datetime.now())+"_" +
                      str(user_id) + '.jpg',
                      Body=requests.get(update.MediaUrl0,
                                        allow_redirects=True).content,
                      ACL='private',
                      # Metadata={
                      #    'x-amz-meta-my-key': 'your-value'
                      # }
                      )

def twilio_write_message(client, update: TwilioUpdate, context, bucket, folder):
    user_id = update.WaId
    name = update.ProfileName
    message = update.Body

    s3_client.put_object(Bucket=bucket,
                      Key=folder + "/" +
                      str(datetime.now())+"_"+str(user_id) + '.txt',
                      Body=message,
                      ACL='private',
                      # Metadata={
                      #    'x-amz-meta-my-key': 'your-value'
                      # }
                      )


def twilio_write_voice(client, update: TwilioUpdate, context, bucket, folder):
    user_id = update.WaId
    name = update.ProfileName

    s3_client.put_object(Bucket=bucket,
                      Key=folder + "/" +
                      str(datetime.now())+"_"+str(user_id) + '.mp3',
                      Body=requests.get(update.MediaUrl0,
                                        allow_redirects=True).content,
                      ACL='private',
                      # Metadata={
                      #    'x-amz-meta-my-key': 'your-value'
                      # }
                      )

def twilio_write(client, update: TwilioUpdate, context, bucket, folder):
    if update.MediaContentType0 and update.MediaContentType0.startswith("audio"):
        twilio_write_voice(client, update, context, bucket, folder)
    elif update.MediaContentType0 and update.MediaContentType0.startswith("image"):
        twilio_write_photo(client, update, context, bucket, folder)
    elif update.Body != "":
        twilio_write_message(client, update, context, bucket, folder)
    else:
        raise NotImplementedError("This type of update can not be saved")


twilio_action_functions = {"write_photo": twilio_write_photo,
                             "write_message": twilio_write_message,
                             "write_voice": twilio_write_voice,
                             "write": twilio_write
                             }



import mongoengine
import datetime

from configparser import ConfigParser

config = ConfigParser()
config.read('config.ini')

dbname = config["bot"]["bot_name"]
class WhatsAppUser(mongoengine.Document):
    ProfileName = mongoengine.StringField(required=True)
    WaId = mongoengine.StringField(required=True, max_length=50)
    entry_time = mongoengine.DateTimeField(default=datetime.datetime.utcnow)
    meta = {'db_alias': dbname}

class WhatsAppInteraction(mongoengine.Document):
    user = mongoengine.ReferenceField(WhatsAppUser)
    text = mongoengine.StringField()
    WaId = mongoengine.StringField()
    ProfileName = mongoengine.StringField()
    date = mongoengine.DateTimeField(default=datetime.datetime.utcnow)
    meta = {'db_alias': dbname}

class WhatsAppUserContextState(mongoengine.Document):
    WaId = mongoengine.StringField(primary_key=True)
    context = mongoengine.DictField()
    state = mongoengine.StringField()
    meta = {'db_alias': dbname}


class WhatsAppUpdate:

    def __init__(self, object, entry) -> None:
        self.object = object
        self.entry = [WhatsAppEntry(**e) for e in entry]

    def get_from(self):
        return self.entry[0].changes[0].value.contacts[0].wa_id

    def get_message(self):
        return self.entry[0].changes[0].value.messages[0]

    def get_message_text(self):
        message = self.entry[0].changes[0].value.messages[0]
        if message._type=="text":
           return message.text
        elif message._type=="interactive":
            if message.interactive["type"] == "button_reply":
                print("Interactive Message Text: {}".format(message.interactive["button_reply"]["title"]))
                return message.interactive["button_reply"]["title"]
            else:
                print("There is no message text defined for this interactive type: {}".format(message.interactive["type"]))
                return ""
        else:
            print("There is no message text defined for this message type: {}".format(message._type))

    def get_delivery(self):
        if self.entry[0].changes[0].value.statuses:
            return self.entry[0].changes[0].value.statuses[0].status


class WhatsAppEntry:
    def __init__(self, id, changes) -> None:
        self.id = id
        self.changes = [WhatsAppChange(**change) for change in changes]


class WhatsAppChange:
    def __init__(self, value, field):
        self.value = WhatsAppValue(**value)
        self.field = field


class WhatsAppValue:
    def __init__(self, metadata, contacts=[], errors=[], messaging_product="whatsapp", messages=[], statuses=[]):
        self.contacts = [WhatsAppContact(**contact) for contact in contacts]
        self.errors = [WhatsAppError(**error) for error in errors]
        self.messaging_product = messaging_product
        messages_list = []
        for message in messages:
            message["_from"] = message["from"]
            del message["from"]
            messages_list.append(message)
        self.messages = [WhatsAppMessage(**message) for message in messages_list]
        self.metadata = WhatsAppMetadata(**metadata)
        self.statuses = [WhatsAppStatus(**status) for status in statuses]


class WhatsAppContact:
    def __init__(self, wa_id, profile) -> None:
        self.wa_id = wa_id
        self.profile_name = profile["name"]


class WhatsAppError:
    def __init__(self, code, title) -> None:
        self.code = code
        self.title = title


class WhatsAppMessage:
    def __init__(self, _from, id, timestamp, type, audio=None, button=None, context=None, document=None, errors=None, identity=None, image=None, interactive=None, order=None, referral=None, sticker=None, system=None, text=None, video=None) -> None:
        if type== "audio":
            self.audio = WhatsAppAudio(**audio)
        elif type=="button":
            self.button = WhatsAppButton(**button)
        elif type=="text":
            self.text = text["body"]
        self.context = context
        self.document = document
        self.errors = errors
        self._from = _from
        self.id = id
        self.identity = identity
        self.image = image
        self.interactive = interactive
        self.order = order
        self.referral = referral
        self.sticker = sticker
        self.system = system
        self.timestamp = timestamp
        self._type = type
        self.video = video


class WhatsAppAudio:
    def __init__(self, id, mime_type) -> None:
        self.id = id
        self.mime_type = mime_type


class WhatsAppButton:
    def __init__(self, payload, text) -> None:
        self.payload = payload
        self.text = text


class WhatsAppMetadata:
    def __init__(self, display_phone_number, phone_number_id) -> None:
        self.display_phone_number = display_phone_number
        self.phone_number_id = phone_number_id


class WhatsAppStatus:
    def __init__(self, id, recipient_id, status, timestamp, conversation=None, pricing=None) -> None:
        if conversation:
            self.conversation = WhatsAppConversation(**conversation)
        self.id = id
        if pricing:
            self.pricing_category = pricing["category"]
            self.pricing_model = pricing["pricing_model"]
        self.recipient_id = recipient_id
        self.status = status
        self.timestamp = timestamp


class WhatsAppConversation:
    def __init__(self, id, origin, expiration_timestamp=None) -> None:
        self.id = id
        self.origin_type = origin["type"]
        if expiration_timestamp:
            self.expiration_timestamp = expiration_timestamp



from digitalguide.pattern import JAHRESZAHL_PATTERN, KOMMAZAHL_PATTERN

from digitalguide.twilio.TwilioUpdate import TwilioUpdate
import re
import warnings

def twilio_eval_jahreszahl(client, update: TwilioUpdate, context, echter_wert, richtig_text, vorher_singular_text, vorher_plural_text, spaeter_singular_text, spaeter_plural_text):
    echter_wert = int(echter_wert)
    schaetzung = int(re.findall(JAHRESZAHL_PATTERN, update.Body)[0])
    if schaetzung == echter_wert:
        client.messages.create(
            body=richtig_text,
            from_=update.To,
            to=update.From
        )

    differenz = schaetzung - echter_wert
    if differenz == -1:
        client.messages.create(
            body=spaeter_singular_text,
            from_=update.To,
            to=update.From
        )

    elif differenz < -1:
        client.messages.create(
            body=spaeter_plural_text.format(abs(differenz)),
            from_=update.To,
            to=update.From
        )

    elif differenz == 1:
        client.messages.create(
            body=vorher_singular_text,
            from_=update.To,
            to=update.From
        )

    elif differenz > 1:
        client.messages.create(
            body=vorher_plural_text.format(abs(differenz)),
            from_=update.To,
            to=update.From
        )


def twilio_eval_prozentzahl(client, update: TwilioUpdate, context, echter_wert, richtig_text, falsch_text):
    warnings.warn("twilio_eval_prozentzahl is deprecated", DeprecationWarning)
    return twilio_eval_kommazahl(update, context, echter_wert, richtig_text, falsch_text, falsch_text, falsch_text, falsch_text)

def twilio_eval_kommazahl(client, update: TwilioUpdate, context, echter_wert, richtig_text, vorher_singular_text, vorher_plural_text, spaeter_singular_text, spaeter_plural_text):
    echter_wert = float(echter_wert)
    match = re.search(
        KOMMAZAHL_PATTERN, update.Body)

    if match:
        if match.group('vorkomma'):
            schaetzung = int(match.group('vorkomma'))
        else:
            schaetzung = 0

        if match.group('nachkomma'):
            schaetzung += float("0."+match.group('nachkomma'))

        if schaetzung == echter_wert:
            client.messages.create(
                body=richtig_text,
                from_=update.To,
                to=update.From
            )

        differenz = schaetzung - echter_wert
        if differenz == -1:
            client.messages.create(
                body=spaeter_singular_text,
                from_=update.To,
                to=update.From
            )

        elif differenz < -1 or (-1 < differenz < 0):
            client.messages.create(
                body=spaeter_plural_text.format(abs(differenz)),
                from_=update.To,
                to=update.From
            )

        elif differenz == 1:
            client.messages.create(
                body=vorher_singular_text,
                from_=update.To,
                to=update.From
            )

        elif differenz > 1 or (0 < differenz < 1):
            client.messages.create(
                body=vorher_plural_text.format(abs(differenz)),
                from_=update.To,
                to=update.From
            )

def twilio_eval_laenge(client, update: TwilioUpdate, context, echter_wert, richtig_text, vorher_singular_text, vorher_plural_text, spaeter_singular_text, spaeter_plural_text):
    warnings.warn("twilio_eval_laenge is deprecated", DeprecationWarning)
    return twilio_eval_kommazahl(client, update, context, echter_wert, richtig_text, vorher_singular_text, vorher_plural_text, spaeter_singular_text, spaeter_plural_text)

twilio_action_functions = {"eval_jahreszahl": twilio_eval_jahreszahl,
                             "eval_prozentzahl": twilio_eval_prozentzahl,
                             "eval_kommazahl": twilio_eval_kommazahl,
                             "eval_laenge": twilio_eval_laenge,
                             }




from telegram import (ReplyKeyboardRemove, Update)
from telegram.ext import (CallbackContext)

from digitalguide.twilio.TwilioUpdate import TwilioUpdate

def telegram_loop_list(update: Update, context: CallbackContext, key, value, doppelte_antwort):
    if not key in context.user_data:
        context.user_data[key] = []

    if value in context.user_data[key]:
        update.message.reply_text(doppelte_antwort,
                                    reply_markup=ReplyKeyboardRemove())
        return "{}_FRAGE".format(key.upper())
    else:
        context.user_data[key].append(value)

def telegram_loop_list_fertig(update: Update, context: CallbackContext, key, answer_id_list, fertig_antwort):
    if set(answer_id_list).issubset(set(context.user_data[key])):
        update.message.reply_text(fertig_antwort,
                                    reply_markup=ReplyKeyboardRemove())


def twilio_loop_list(client, update: TwilioUpdate, context,  key, value, doppelte_antwort):
    if not key in context:
        context[key] = []

    if value in context[key]:
        client.messages.create(
            body=doppelte_antwort,
            from_=update.To,
            to=update.From
        )
        return "{}_FRAGE".format(key.upper())
    else:
        context[key].append(value)

def twilio_loop_list_fertig(client, update, context, key, answer_id_list, fertig_antwort):
    if set(answer_id_list).issubset(set(context[key])):
        client.messages.create(
            body=fertig_antwort,
            from_=update.To,
            to=update.From
        )

def telegram_eval_list(update: Update, context: CallbackContext, answer_id_name_list, poi, response_text):
    if not poi in context.user_data:
        context.user_data[poi] = []

    response_text += "\n"

    for id, name in answer_id_name_list:
        if id in context.user_data[poi]:
            response_text += "✅ {}\n".format(name)
        else:
            response_text += "◽ {}\n".format(name)

    update.message.reply_text(response_text,
                              reply_markup=ReplyKeyboardRemove())


def twilio_eval_list(client, update: TwilioUpdate, context, answer_id_name_list, poi, response_text):
    if not poi in context:
        context[poi] = []

    response_text += "\n"

    for id, name in answer_id_name_list:
        if id in context[poi]:
            response_text += "✅ {}\n".format(name)
        else:
            response_text += "◽ {}\n".format(name)

    client.messages.create(
        body=response_text,
        from_=update.To,
        to=update.From
    )


telegram_action_functions = {"loop_list": telegram_loop_list,
                             "eval_list": telegram_eval_list,
                             "loop_list_fertig": telegram_loop_list_fertig
                             }

twilio_action_functions = {"loop_list": twilio_loop_list,
                             "eval_list": twilio_eval_list,
                             "loop_list_fertig": twilio_loop_list_fertig
                             }



from setuptools import find_packages, setup

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

test_deps = ['pytest',
             'flake8',
             "google-api-python-client",
             "google-auth-httplib2",
             "google-auth-oauthlib",
             "flask"]

twilio_deps = ["twilio",
               "quart",
               "hypercorn==0.13.2",
               "redis"
               ]

whatsapp_deps = ["quart",
                 "hypercorn==0.13.2",
                 "heyoo==0.0.4"]

extras = {
    'test': test_deps,
    'twilio': twilio_deps,
    'whatsapp': whatsapp_deps
}

setup(
    name='digitalguide',
    packages=find_packages(),
    version='0.0.327',
    description='A Python Library to write digital guides for telegram',
    author='Soeren Etler',
    license='MIT',
    install_requires=["python-telegram-bot",
                      "ctparse",
                      "boto3",
                      "pymongo[srv]>= 3.4, < 4.0",
                      "mongoengine",
                      "Pillow",
                      "PyYAML",
                      "requests"],
    setup_requires=['pytest-runner'],
    tests_require=test_deps,
    extras_require=extras,
    test_suite='tests',
    long_description=long_description,
    long_description_content_type='text/markdown'
)



import asyncio
from collections import defaultdict
from digitalguide.twilio.TwilioUpdate import TwilioUpdate
import yaml
import redis

import time

from configparser import ConfigParser
config = ConfigParser()
config.read('config.ini')

import os

def read_action_yaml(filename, action_functions={}):
    with open(filename) as file:
        yaml_dict = yaml.load(file, Loader=yaml.FullLoader)

    actions_dict = {}

    for key, value in yaml_dict.items():
        actions_dict[key] = Action(value, action_functions=action_functions)

    return actions_dict

redis_url=os.environ.get("REDIS_URL")
r_unsend_messages = redis.from_url(redis_url, decode_responses=True)

class Action():
    def __init__(self, actions, action_functions={}, log_user=True, log_interaction=True):
        self.actions = actions
        self.action_functions = action_functions
        self.log_user = log_user
        self.log_interaction = log_interaction
        if log_user or log_interaction:
            import mongoengine
            dbname = config["bot"]["bot_name"]
            if os.getenv('DATABASE_CERT', None):
                # This can be remove in the future
                with open("ca-certificate.crt", "w") as text_file:
                    text_file.write(os.getenv('DATABASE_CERT'))
                mongoengine.connect(alias=dbname, host="mongodb+srv://" + os.getenv("DATABASE_USERNAME")+":" + os.getenv("DATABASE_PASSWORD") +
                                    "@" + os.getenv("DATABASE_HOST") + "/"+dbname+"?authSource=admin&tls=true&tlsCAFile=ca-certificate.crt")
            else:
                mongoengine.connect(alias=dbname, host="mongodb+srv://" + os.getenv("DATABASE_USERNAME")+":" + os.getenv("DATABASE_PASSWORD") +
                                    "@" + os.getenv("DATABASE_HOST") + "/"+dbname+"?authSource=admin&tls=true")
            


    async def __call__(self, client, update: TwilioUpdate, context):
        if self.log_user and not ("user_id" in context.keys()):
            from digitalguide.twilio.db_objects import TwilioUser
            db_user = TwilioUser(ProfileName=update.ProfileName,
                           WaId=update.WaId)
            db_user.save()
            context["user_id"] = db_user

        if self.log_interaction:
            from digitalguide.twilio.db_objects import TwilioInteraction

            TwilioInteraction(user=context["user_id"],
                            Latitude = update.Latitude,
                            Longitude = update.Longitude,
                            SmsMessageSid = update.SmsMessageSid,
                            NumMedia = update.NumMedia,
                            ProfileName = update.ProfileName,
                            SmsSid = update.SmsSid,
                            WaId = update.WaId,
                            SmsStatus = update.SmsStatus,
                            Body = update.Body,
                            To = update.To,
                            NumSegments = update.NumSegments,
                            MessageSid = update.MessageSid,
                            AccountSid = update.AccountSid,
                            From = update.From,
                            ApiVersion = update.ApiVersion,
                            MediaContentType0 = update.MediaContentType0,
                            MediaUrl0 = update.MediaUrl0,
                            MessagingServiceSid = update.MessagingServiceSid,
                            ReferralNumMedia = update.ReferralNumMedia
                            ).save()




        for item in self.actions:
            if item["type"] == "return":
                return item["state"]

            max_wait_itteration = 30
            for i in range(max_wait_itteration):
                print("unsend messages iterration {}: {}".format(i, r_unsend_messages.smembers(update.From)))
                if not r_unsend_messages.smembers(update.From):
                    print("BREAK")
                    break
                await asyncio.sleep(.1)
            else:
                print("Previous messages were not send within {} ms.".format(max_wait_itteration))

            if item["type"] == "message":
                message = client.messages.create(
                    body=item["text"].format(
                        **{"profileName": update.ProfileName, "echo": update.Body, **context}),
                    from_=update.To,
                    to=update.From
                )
                print("New unsend message: {}".format(message.sid))
                r_unsend_messages.sadd(update.From, message.sid)
                
            elif item["type"] == "venue":
                message = client.messages.create(
                    body=item["title"],
                    persistent_action=['geo:{},{}|{}'.format(
                        item["latitude"], item["longitude"], item["address"])],
                    from_=update.To,
                    to=update.From
                )
                r_unsend_messages.sadd(update.From, message.sid)
            elif item["type"] == "photo":
                message = client.messages.create(
                    media_url=item["url"],
                    from_=update.To,
                    to=update.From
                )
                r_unsend_messages.sadd(update.From, message.sid)
            elif item["type"] == "video":
                message = client.messages.create(
                    media_url=item["url"],
                    from_=update.To,
                    to=update.From
                )
                r_unsend_messages.sadd(update.From, message.sid)
            elif item["type"] == "media_group":
                message = client.messages.create(
                    media_url=item["urls"],
                    from_=update.To,
                    to=update.From
                )
                r_unsend_messages.sadd(update.From, message.sid)
            elif item["type"] == "audio" or item["type"] == "voice":
                message = client.messages.create(
                    media_url=[item["url"]],
                    from_=update.To,
                    to=update.From
                )
                r_unsend_messages.sadd(update.From, message.sid)
            elif item["type"] == "poll":
                message = item["question"] + "\n"
                for option in item["options"]:
                    message += option + "\n"
                message = client.messages.create(
                    body=message,
                    from_=update.To,
                    to=update.From
                )
                r_unsend_messages.sadd(update.From, message.sid)

            elif item["type"] == "function":
                arguments = {i: item[i]
                             for i in item if i != 'type' and i != 'func'}
                self.action_functions[item["func"]](
                    client, update, context, **arguments)


from digitalguide.twilio.TwilioUpdate import TwilioUpdate

def twilio_default_name(client, update: TwilioUpdate, context):
    context["name"] = update.ProfileName

def twilio_save_text_to_context(client, update: TwilioUpdate, context, key):
    context[key] = update.Body

def twilio_save_value_to_context(client, update: TwilioUpdate, context, key, value):
    context[key] = value

twilio_action_functions = {"default_name": twilio_default_name,
                             "save_text_to_context": twilio_save_text_to_context,
                             "save_value_to_context": twilio_save_value_to_context
                             }
